"""
All files in this schema directory and below were copied from the freely distributed BIDS specification repository.
The original files are released with a Creative Commons Attribution 4.0 International Public (CC-BY-4.0) License. 

For more information see:
https://github.com/bids-standard/bids-specification
https://github.com/bids-standard/bids-specification/blob/master/LICENSE
"""
